from unittest import TestCase

import functools

from rapt2.treebrd.condition_node import (
    BinaryConditionNode,
    BinaryConditionalOperator,
    IdentityConditionNode,
)
from rapt2.treebrd.errors import RelationReferenceError
from rapt2.treebrd.grammars import ExtendedGrammar
from rapt2.treebrd.node import (
    RelationNode,
    ProjectNode,
    CrossJoinNode,
    NaturalJoinNode,
    ThetaJoinNode,
)
from rapt2.treebrd.schema import Schema
from rapt2.treebrd.treebrd import TreeBRD


class TreeBRDTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.schema = None

    @classmethod
    def create_build_function(cls, schema):
        builder = TreeBRD(ExtendedGrammar())
        return functools.partial(builder.build, schema=schema)


class TestRelation(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "letters": ["position", "value"],
            "numbers": ["value", "prime"],
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))

    def test_build_when_instring_is_single_relation(self):
        forest = self.build("letters;")
        self.assertEqual(1, len(forest))
        expected = RelationNode("letters", self.schema)
        self.assertEqual(expected, forest[0])

    def test_build_when_instring_has_multiple_relations(self):
        instring = "numbers; letters;"
        forest = self.build(instring)
        self.assertEqual(2, len(forest))
        first = RelationNode("numbers", self.schema)
        second = RelationNode("letters", self.schema)
        self.assertEqual(first, forest[0])
        self.assertEqual(second, forest[1])


class TestProject(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "magic_wand": [
                "owner",
                "manufacturer",
                "wood",
                "core",
                "length",
                "rigidity",
            ]
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))

    def test_project_with_single_attr(self):
        instring = r"\project_{owner} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, ["owner"])
        self.assertEqual(expected, forest[0])

    def test_project_with_two_attr(self):
        instring = r"\project_{owner, wood} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, ["owner", "wood"])
        self.assertEqual(expected, forest[0])

    def test_project_with_all_but_one_attr(self):
        attr = self.schema.get_attributes("magic_wand")
        attr.remove("rigidity")
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])

    def test_project_with_all_attr(self):
        attr = self.schema.get_attributes("magic_wand")
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])

    def test_project_with_all_attr_shuffled(self):
        attr = self.schema.get_attributes("magic_wand")
        attr.sort()
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])


class JoinTestCase(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "alpha": ["a1"],
            "beta": ["b1", "b2"],
            "gamma": ["c1", "c2", "c3"],
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))


class TestJoins(JoinTestCase):
    def test_join_with_natural_join(self):
        instring = "alpha \\join beta \\natural_join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        expected = NaturalJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_natural_join_with_join(self):
        instring = "alpha \\natural_join beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = NaturalJoinNode(left, middle)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_join_with_theta_join(self):
        instring = "alpha \\join beta \\theta_join_{a1 = c1} gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("c1"),
        )
        expected = ThetaJoinNode(intermediate, right, condition)
        self.assertEqual(expected, forest[0])

    def test_theta_join_with_join(self):
        instring = "alpha \\theta_join_{a1 = b1} beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        intermediate = ThetaJoinNode(left, middle, condition)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])


class TestCrossJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\join beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        expected = CrossJoinNode(left, right)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\join beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, CrossJoinNode, left, right)


class TestThetaJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\theta_join_{a1 = b1} beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        expected = ThetaJoinNode(left, right, condition)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\theta_join_{a1 = b1} beta \\theta_join_{a1 = b1} gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        intermediate = ThetaJoinNode(left, middle, condition)
        expected = ThetaJoinNode(intermediate, right, condition)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, ThetaJoinNode, left, right, "a1=5")


class TestNaturalJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\natural_join beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        expected = NaturalJoinNode(left, right)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\natural_join beta \\natural_join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = NaturalJoinNode(left, middle)
        expected = NaturalJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, NaturalJoinNode, left, right)



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestAssignment(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_assign_relation(self):
        expression = "asts := astronauts;"
        expected = [[["asts"], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_expression(self):
        expression = "ast := \\project_{name} astronauts;"
        expected = [[["ast"], ":=", ["\\project", ["name"], ["astronauts"]]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_multiple_assign_expressions(self):
        expression = (
            "names := \\project_{name} astronauts;"
            "species := \\project_{species} astronauts;"
        )
        expected = [
            [["names"], ":=", ["\\project", ["name"], ["astronauts"]]],
            [["species"], ":=", ["\\project", ["species"], ["astronauts"]]],
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_attributes(self):
        expression = "ast(age, happiness) := astronauts;"
        expected = [[["ast", ["age", "happiness"]], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_attributes_spacing(self):
        expression = "ast ( age , happiness ) := astronauts;"
        expected = [[["ast", ["age", "happiness"]], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_no_expressions(self):
        expression = ":=;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_right(self):
        expression = "A := ;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_no_left(self):
        expression = ":= A;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_no_attributes(self):
        expression = "ast() := A;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )



# Relational algebra latex operators

PROJECT_OP = "\\pi"
RENAME_OP = "\\rho"
SELECT_OP = "\\sigma"
ASSIGN_OP = ":="

JOIN_OP = "\\times"
THETA_JOIN_OP = "\\times"
NATURAL_JOIN_OP = "\\bowtie"
FULL_OUTER_JOIN_OP = "\\fullouterjoin"
LEFT_OUTER_JOIN_OP = "\\leftouterjoin"
RIGHT_OUTER_JOIN_OP = "\\rightouterjoin"

DIFFERENCE_OP = "-"
UNION_OP = "\\cup"
INTERSECT_OP = "\\cap"



from unittest import TestCase
from rapt2.treebrd.condition_node import (
    IdentityConditionNode,
    UnaryConditionNode,
    UnaryConditionalOperator,
    BinaryConditionalOperator,
    BinaryConditionNode,
)
from rapt2.treebrd.schema import Schema


class ConditionNodeTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def setUp(self):
        self.schema = Schema(
            {
                "alpha": ["a1"],
                "beta": ["b1", "b2"],
                "gamma": ["c1", "c2", "c3"],
            }
        )


class TestConditionNode(ConditionNodeTestCase):
    def test_identity_condition_node(self):
        node = IdentityConditionNode("alpha.a1")
        self.assertEqual(node.to_latex(), "alpha.a1")
        self.assertEqual(node.to_sql(), "alpha.a1")
        self.assertEqual(node.attribute_references(), ["alpha.a1"])

    def test_unary_condition_node(self):
        child = IdentityConditionNode("alpha.a1")
        node = UnaryConditionNode(UnaryConditionalOperator.NOT, child)
        self.assertEqual(node.to_latex(), "\\neg alpha.a1")
        self.assertEqual(node.to_sql(), "NOT alpha.a1")
        self.assertEqual(node.attribute_references(), ["alpha.a1"])

    def test_binary_condition_node(self):
        left = IdentityConditionNode("alpha.a1")
        right = IdentityConditionNode("beta.b1")
        node = BinaryConditionNode(BinaryConditionalOperator.AND, left, right)
        self.assertEqual(node.to_latex(), "(alpha.a1 \\land beta.b1)")
        self.assertEqual(node.to_sql(), "(alpha.a1 AND beta.b1)")
        self.assertEqual(node.attribute_references(), ["alpha.a1", "beta.b1"])



from .constants import *
from ...treebrd.node import Operator

latex_operator = {
    Operator.select: SELECT_OP,
    Operator.project: PROJECT_OP,
    Operator.rename: RENAME_OP,
    Operator.assign: ASSIGN_OP,
    Operator.cross_join: JOIN_OP,
    Operator.natural_join: NATURAL_JOIN_OP,
    Operator.theta_join: THETA_JOIN_OP,
    Operator.full_outer_join: FULL_OUTER_JOIN_OP,
    Operator.left_outer_join: LEFT_OUTER_JOIN_OP,
    Operator.right_outer_join: RIGHT_OUTER_JOIN_OP,
    Operator.union: UNION_OP,
    Operator.difference: DIFFERENCE_OP,
    Operator.intersect: INTERSECT_OP,
}



class Syntax:
    def __init__(self, **kwargs):
        # First initialize a default syntax.

        # General tokens.
        self.terminator = ";"
        self.delim = ","
        self.params_start = "_{"
        self.params_stop = "}"
        self.paren_left = "("
        self.paren_right = ")"

        # Logical tokens.
        self.not_op = "not"
        self.and_op = "and"
        self.or_op = "or"

        # Comparison operators.
        self.equal_op = "="
        self.not_equal_op = "!="
        self.not_equal_alt_op = "<>"
        self.less_than_op = "<"
        self.less_than_equal_op = "<="
        self.greater_than_op = ">"
        self.greater_than_equal_op = ">="

        # Relational algebra operators.
        self.project_op = "\\project"
        self.rename_op = "\\rename"
        self.select_op = "\\select"
        self.assign_op = ":="
        self.join_op = "\\join"
        self.theta_join_op = "\\theta_join"
        self.natural_join_op = "\\natural_join"
        self.full_outer_join_op = "\\full_outer_join"
        self.left_outer_join_op = "\\left_outer_join"
        self.right_outer_join_op = "\\right_outer_join"
        self.difference_op = "\\difference"
        self.union_op = "\\union"
        self.intersect_op = "\\intersect"

        # Defined operator.
        self.defined_op = "defined"

        # Now set any user defined syntax.
        for key in kwargs:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])



from pyparsing import ParseResults
from rapt2.treebrd.condition_node import (
    IdentityConditionNode,
    BinaryConditionNode,
    BinaryConditionalOperator,
    ConditionNode,
    UnaryConditionNode,
    UnaryConditionalOperator,
)
from .schema import Schema
from .node import (
    SelectNode,
    ProjectNode,
    RenameNode,
    AssignNode,
    CrossJoinNode,
    NaturalJoinNode,
    UnionNode,
    DifferenceNode,
    IntersectNode,
    ThetaJoinNode,
    FullOuterJoinNode,
    LeftOuterJoinNode,
    RightOuterJoinNode,
    RelationNode,
)
from .grammars.proto_grammar import ProtoGrammar


class TreeBRD:
    """
    A Tree Builder for Relational (Algebra) Data. TreeBRD is a factory class
    that builds forests of relational algebra syntax trees. STARBuilder
    """

    grammar: ProtoGrammar

    def __init__(self, grammar: ProtoGrammar):
        self.grammar = grammar

    def build(self, instring, schema):
        ra = self.grammar.parse(instring)
        _schema = Schema(schema)
        return [self.to_node(statement, _schema) for statement in ra[:]]

    def to_node(self, exp: ParseResults, schema):
        """
        Return a Node that is the root of the parse tree for the the specified
        expression.

        :param exp: A list that represents a relational algebra expression.
        Assumes that this list was generated by pyparsing.
        :param schema: A dictionary of relation names to attribute names used
        for verification and generating attributes.
        :return: A Node.
        """

        # A relation node.
        if len(exp) == 1 and isinstance(exp[0], str):
            name = exp[0]
            node = RelationNode(name=name, schema=schema)

        # An expression.
        elif len(exp) == 1 and isinstance(exp[0], ParseResults):
            node = self.to_node(exp[0], schema)

        # Unary operators.
        elif isinstance(exp[0], str) and self.grammar.is_unary(exp[0]):
            child = self.to_node(exp[2:], schema)
            node = self.create_unary_node(
                operator=exp[0], child=child, param=exp[1], schema=schema
            )

        # Assignment.
        elif exp[1] is self.grammar.syntax.assign_op:
            child = self.to_node(exp[2:], schema)
            node = self.create_unary_node(
                operator=exp[1], child=child, param=exp[0], schema=schema
            )

        # Binary operators.
        elif self.grammar.is_binary(exp[1]):
            # Pyparsing will put different operators with the same precedence
            # in the same list. This can be a problem when we mix operators with
            # and without parameters (for example join). We avoid this below
            # and build from right to left, to create the correct syntax tree.
            if isinstance(exp[-2], str):
                # Operator without parameters
                op_pos = -2
                param = None
            else:
                op_pos = -3
                param = exp[-2]

            operator = exp[op_pos]
            left = self.to_node(exp[:op_pos], schema)
            right = self.to_node(exp[-1], schema)
            node = self.create_binary_node(
                operator=operator, left=left, right=right, param=param
            )

        else:
            raise ValueError

        return node

    def create_condition_node(self, conditions: ParseResults) -> ConditionNode:
        if isinstance(conditions, str):
            # Attribute Condition Node
            return IdentityConditionNode(conditions)

        elif len(conditions) == 2:
            # Unary Condition Node
            op, right = conditions
            right_node = self.create_condition_node(right)

            return UnaryConditionNode(
                op=UnaryConditionalOperator.from_syntax(self.grammar.syntax, op),
                child=right_node,
            )

        elif len(conditions) == 3:
            # Binary Condition Node
            left, op, right = conditions
            left_node = self.create_condition_node(left)
            right_node = self.create_condition_node(right)

            return BinaryConditionNode(
                op=BinaryConditionalOperator.from_syntax(self.grammar.syntax, op),
                left=left_node,
                right=right_node,
            )

        raise ValueError

    def create_unary_node(
        self, operator, child, param: ParseResults | None = None, schema=None
    ):
        """
        Return a Unary Node whose type depends on the specified operator.

        :param schema:
        :param child:
        :param operator: A relational algebra operator (see constants.py)
        :param param: A list of parameters for the operator.
        :return: A Unary Node.
        """

        if operator == self.grammar.syntax.select_op:
            condition = self.create_condition_node(param[0])
            node = SelectNode(child, condition)

        elif operator == self.grammar.syntax.project_op:
            node = ProjectNode(child, param)

        elif operator == self.grammar.syntax.rename_op:
            name = None
            attributes = []
            if isinstance(param[0], str):
                name = param.pop(0)
            if param:
                attributes = param[0]
            node = RenameNode(child, name, attributes, schema)

        elif operator == self.grammar.syntax.assign_op:
            name = param[0]
            attributes = [] if len(param) < 2 else param[1]
            node = AssignNode(child, name, attributes, schema)

        else:
            raise ValueError

        return node

    def create_binary_node(
        self, operator, left, right, param: ParseResults | None = None
    ):
        """
        Return a Node whose type depends on the specified operator.

        :param operator: A relational algebra operator (see constants.py)
        :return: A Node.
        """

        # Join operators
        if operator == self.grammar.syntax.join_op:
            node = CrossJoinNode(left, right)

        elif operator == self.grammar.syntax.natural_join_op:
            node = NaturalJoinNode(left, right)

        elif operator == self.grammar.syntax.theta_join_op:
            condition = self.create_condition_node(param[0])
            node = ThetaJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.full_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = FullOuterJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.left_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = LeftOuterJoinNode(left, right, condition)

        elif operator == self.grammar.syntax.right_outer_join_op:
            condition = self.create_condition_node(param[0])
            node = RightOuterJoinNode(left, right, condition)

        # Set operators
        elif operator == self.grammar.syntax.union_op:
            node = UnionNode(left, right)

        elif operator == self.grammar.syntax.difference_op:
            node = DifferenceNode(left, right)

        elif operator == self.grammar.syntax.intersect_op:
            node = IntersectNode(left, right)

        else:
            raise ValueError

        return node



from ...treebrd.node import Operator
from ..base_translator import BaseTranslator


class SQLQuery:
    """
    Structure defining the building blocks of a SQL query.
    """

    def __init__(self, select_block, from_block, where_block=""):
        self.prefix = ""
        self.select_block = select_block
        self.from_block = from_block
        self.where_block = where_block

    @property
    def _basic_query(self):
        if self.select_block:
            return "{prefix}SELECT {select} FROM {relation}"
        else:
            return "{prefix}{relation}"

    @property
    def _sql_query_skeleton(self):
        sql = self._basic_query
        if self.where_block:
            sql += " WHERE {conditions}"
        return sql

    def to_sql(self):
        """
        Construct a SQL query based on the stored blocks.
        :return: a SQL query
        """
        return self._sql_query_skeleton.format(
            prefix=self.prefix,
            select=self.select_block,
            relation=self.from_block,
            conditions=self.where_block,
        )


class SQLSetQuery(SQLQuery):
    """
    Structure defining the building blocks of a SQL query with set semantics.
    """

    @property
    def _basic_query(self):
        return "{prefix}SELECT DISTINCT {select} FROM {relation}"


class Translator(BaseTranslator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a SQL statement using bag semantics.
    """

    query = SQLQuery

    @classmethod
    def _get_temp_name(cls, node):
        return node.name or "_{}".format(id(node))

    @classmethod
    def _get_sql_operator(cls, node):
        operators = {
            Operator.union: "UNION",
            Operator.difference: "EXCEPT",
            Operator.intersect: "INTERSECT",
            Operator.cross_join: "CROSS JOIN",
            Operator.theta_join: "JOIN",
            Operator.natural_join: "NATURAL JOIN",
            Operator.full_outer_join: "FULL OUTER JOIN",
            Operator.left_outer_join: "LEFT OUTER JOIN",
            Operator.right_outer_join: "RIGHT OUTER JOIN",
        }
        return operators[node.operator]

    def relation(self, node):
        """
        Translate a relation node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self.query(select_block=str(node.attributes), from_block=node.name)

    def select(self, node):
        """
        Translate a select node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """

        child_object = self.translate(node.child)
        where_block = node.conditions.to_sql()
        if child_object.where_block:
            where_block = "({0}) AND ({1})".format(
                child_object.where_block, node.conditions.to_sql()
            )
        child_object.where_block = where_block
        if not child_object.select_block:
            child_object.select_block = str(node.attributes)
        return child_object

    def project(self, node):
        """
        Translate a project node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        child_object.select_block = str(node.attributes)
        return child_object

    def rename(self, node):
        """
        Translate a rename node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        from_block = "({child}) AS {name}({attributes})".format(
            child=child_object.to_sql(),
            name=node.name,
            attributes=", ".join(node.attributes.names),
        )
        return self.query(str(node.attributes), from_block=from_block)

    def assign(self, node):
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        child_object = self.translate(node.child)
        child_object.prefix = "CREATE TEMPORARY TABLE {name}({attributes}) AS ".format(
            name=node.name, attributes=", ".join(node.attributes.names)
        )
        return child_object

    def natural_join(self, node):
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def theta_join(self, node):
        """
        Translate an assign node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def cross_join(self, node):
        """
        Translate a cross join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def full_outer_join(self, node):
        """
        Translate a full outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def left_outer_join(self, node):
        """
        Translate a left outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def right_outer_join(self, node):
        """
        Translate a right outer join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._join(node)

    def union(self, node):
        """
        Translate a union node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def intersect(self, node):
        """
        Translate an intersection node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def difference(self, node):
        """
        Translate an difference node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        return self._set_op(node)

    def _join_helper(self, node):
        sobject = self.translate(node)
        if node.operator in {
            Operator.cross_join,
            Operator.natural_join,
            Operator.theta_join,
            Operator.full_outer_join,
            Operator.left_outer_join,
            Operator.right_outer_join,
        }:
            return sobject.from_block
        else:
            return "({subquery}) AS {name}".format(
                subquery=sobject.to_sql(), name=self._get_temp_name(node)
            )

    def _join(self, node):
        """
        Translate a join node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """

        select_block = str(node.attributes)
        from_block = "{left} {operator} {right}".format(
            left=self._join_helper(node.left),
            right=self._join_helper(node.right),
            operator=self._get_sql_operator(node),
        )

        if node.operator in {
            Operator.theta_join,
            Operator.full_outer_join,
            Operator.left_outer_join,
            Operator.right_outer_join,
        }:
            from_block = "{from_block} ON {conditions}".format(
                from_block=from_block, conditions=node.conditions.to_sql()
            )

        return self.query(select_block, from_block, "")

    def _set_op(self, node):
        """
        Translate a set operator node into SQLQuery.
        :param node: a treebrd node
        :return: a SQLQuery object for the tree rooted at node
        """
        select_block = str(node.attributes)
        from_block = "({left} {operator} ALL {right}) AS {name}".format(
            left=self.translate(node.left).to_sql(),
            right=self.translate(node.right).to_sql(),
            operator=self._get_sql_operator(node),
            name=self._get_temp_name(node),
        )
        return self.query(select_block=select_block, from_block=from_block)


class SetTranslator(Translator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a SQL statement using set semantics.
    """

    query = SQLSetQuery

    def _set_op(self, node):
        """
        Translate a set operator node into SQLQuery, using set semantics.
        :param node: a treebrd node
        :return: a SQLSetQuery object for the tree rooted at node
        """
        select_block = str(node.attributes)
        from_block = "({left} {operator} {right}) AS {name}".format(
            left=self.translate(node.left).to_sql(),
            right=self.translate(node.right).to_sql(),
            operator=self._get_sql_operator(node),
            name=self._get_temp_name(node),
        )
        return self.query(select_block=select_block, from_block=from_block)


def translate(root_list, use_bag_semantics=False):
    """
    Translate a list of relational algebra trees into SQL statements.

    :param root_list: a list of tree roots
    :param use_bag_semantics: flag for using relational algebra bag semantics
    :return: a list of SQL statements
    """
    translator = Translator() if use_bag_semantics else SetTranslator()
    return [translator.translate(root).to_sql() for root in root_list]



from rapt2.rapt import Rapt

from rapt2.transformers.qtree.constants import *
from rapt2.treebrd.grammars.extended_grammar import ExtendedGrammar
from tests.transformers.test_transfomer import TestTransformer


class TestQTreeTransformer(TestTransformer):
    def setUp(self):
        self.translate = self.translate_func(Rapt(grammar="Extended Grammar").to_qtree)


class TestRelation(TestQTreeTransformer):
    def test_single_relation(self):
        ra = "alpha;"
        expected = [r"\Tree[.$alpha$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_multiple_relations(self):
        ra = "alpha; beta;"
        expected = [r"\Tree[.$alpha$ ]", r"\Tree[.$beta$ ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestSelect(TestQTreeTransformer):
    def test_simple(self):
        ra = r"\select_{a1=a2} alpha;"
        expected = [r"\Tree[.${}_{{(a1 \eq a2)}}$ [.$alpha$ ] ]".format(SELECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestProject(TestQTreeTransformer):
    def test_simple(self):
        ra = r"\project_{a1, a2, a3} alpha;"
        expected = [r"\Tree[.${}_{{a1, a2, a3}}$ [.$alpha$ ] ]".format(PROJECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRename(TestQTreeTransformer):
    def test_rename_relation(self):
        ra = r"\rename_{Apex} alpha;"
        expected = [r"\Tree[.${}_{{apex(a1, a2, a3)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_attributes(self):
        ra = r"\rename_{(a, b, c)} alpha;"
        expected = [r"\Tree[.${}_{{alpha(a, b, c)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_rename_relation_and_attributes(self):
        ra = r"\rename_{apex(a, b, c)} alpha;"
        expected = [r"\Tree[.${}_{{apex(a, b, c)}}$ [.$alpha$ ] ]".format(RENAME_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestAssignment(TestQTreeTransformer):
    def test_relation(self):
        ra = "new_alpha := alpha;"
        expected = [r"\Tree[.$new_alpha(a1,a2,a3)$ [.$alpha$ ] ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_relation_with_attributes(self):
        ra = "new_alpha(a, b, c) := alpha;"
        expected = [r"\Tree[.$new_alpha(a,b,c)$ [.$alpha$ ] ]"]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestJoin(TestQTreeTransformer):
    def test_relation(self):
        ra = r"alpha \join beta;"
        expected = [r"\Tree[.${}$ [.$alpha$ ] [.$beta$ ] ]".format(JOIN_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestNaturalJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \natural_join beta;"
        expected = [r"\Tree[.${}$ [.$alpha$ ] [.$beta$ ] ]".format(NATURAL_JOIN_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestThetaJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                THETA_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestFullOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \full_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                FULL_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \full_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                FULL_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestLeftOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \left_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                LEFT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \left_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                LEFT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestRightOuterJoin(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_relation(self):
        ra = r"alpha \right_outer_join_{a1 = b1} beta;"
        expected = [
            r"\Tree[.${}_{{(a1 \eq b1)}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                RIGHT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)

    def test_complex_condition(self):
        ra = r"alpha \right_outer_join_{alpha.a1=beta.b1 and b3>50} beta;"
        expected = [
            r"\Tree[.${}_{{((alpha.a1 \eq beta.b1) \land (b3 \gt 50))}}$ [.$alpha$ ] [.$beta$ ] ]".format(
                RIGHT_OUTER_JOIN_OP
            )
        ]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestUnion(TestQTreeTransformer):
    def test_simple(self):
        ra = r"gamma \union gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(UNION_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestIntersect(TestQTreeTransformer):
    grammar = ExtendedGrammar

    def test_simple(self):
        ra = r"gamma \intersect gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(INTERSECT_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)


class TestDifference(TestQTreeTransformer):
    def test_simple(self):
        ra = r"gamma \difference gammatwin;"
        expected = [r"\Tree[.${}$ [.$gamma$ ] [.$gammatwin$ ] ]".format(DIFFERENCE_OP)]
        actual = self.translate(ra)
        self.assertEqual(expected, actual)



from ..base_translator import BaseTranslator
from .operators import latex_operator


class Translator(BaseTranslator):
    """
    A Translator defining the operations for translating a relational algebra
    statement into a latex tree output.
    """

    def relation(self, node):
        """
        Translate a relation node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${}$ ]".format(node.name)

    def select(self, node):
        """
        Translate a select node into a latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        return "[.${op}_{{{conditions}}}$ {child} ]".format(
            op=latex_operator[node.operator],
            conditions=node.conditions.to_latex(),
            child=child,
        )

    def project(self, node):
        """
        Translate a project node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        return "[.${op}_{{{attributes}}}$ {child} ]".format(
            op=latex_operator[node.operator],
            attributes=", ".join(node.attributes.names),
            child=child,
        )

    def rename(self, node):
        """
        Translate a rename node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        attributes = ""
        if node.attributes:
            attributes = "({})".format(", ".join(node.attributes.names))
        return "[.${op}_{{{name}{attributes}}}$ {child} ]".format(
            op=latex_operator[node.operator],
            name=node.name,
            attributes=attributes,
            child=child,
        )

    def assign(self, node):
        """
        Translate an assign node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        child = self.translate(node.child)
        attributes = ""
        if node.attributes:
            attributes = "({})".format(",".join(node.attributes.names))
        return "[.${name}{attributes}$ {child} ]".format(
            name=node.name, attributes=attributes, child=child
        )

    def theta_join(self, node):
        """
        Translate a join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=latex_operator[node.operator],
            conditions=node.conditions.to_latex(),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def cross_join(self, node):
        """
        Translate a cross node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def natural_join(self, node):
        """
        Translate a natural join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def full_outer_join(self, node):
        """
        Translate a full outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=latex_operator[node.operator],
            conditions=node.conditions.to_latex(),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def left_outer_join(self, node):
        """
        Translate a left outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=latex_operator[node.operator],
            conditions=node.conditions.to_latex(),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def right_outer_join(self, node):
        """
        Translate a right outer join node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}_{{{conditions}}}$ {left} {right} ]".format(
            op=latex_operator[node.operator],
            conditions=node.conditions.to_latex(),
            left=self.translate(node.left),
            right=self.translate(node.right),
        )

    def union(self, node):
        """
        Translate a union node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def difference(self, node):
        """
        Translate a difference node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def intersect(self, node):
        """
        Translate an intersect node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return self._binary(node)

    def _binary(self, node):
        """
        Translate a binary node into latex qtree node.
        :param node: a treebrd node
        :return: a qtree subtree rooted at the node
        """
        return "[.${op}$ {left} {right} ]".format(
            op=latex_operator[node.operator],
            left=self.translate(node.left),
            right=self.translate(node.right),
        )


def translate(roots):
    """
    Translate a treebrd tree rooted at root into latex tree.
    :param root: a treebrd node
    :return:  a string representing a latex qtree rooted at root
    """
    return ["\\Tree{root}".format(root=Translator().translate(root)) for root in roots]



import unittest
from pyparsing import ParseException
from rapt2.treebrd.grammars.condition_grammar import (
    ConditionGrammar,
    get_attribute_references,
)
from tests.treebrd.grammars.grammar_test_case import GrammarTestCase


class TestConditionRules(GrammarTestCase):
    def setUp(self):
        self.parser = ConditionGrammar()

    def test_comparator(self):
        parse = self.parse_function(self.parser.comparator_op)

        self.assertEqual(parse("="), ["="])
        self.assertEqual(parse("!="), ["!="])
        self.assertEqual(parse("<>"), ["<>"])
        self.assertEqual(parse("<"), ["<"])
        self.assertEqual(parse("<="), ["<="])
        self.assertEqual(parse(">"), [">"])
        self.assertEqual(parse(">="), [">="])

    def test_comparator_exp(self):
        parse = self.parse_function(self.parser.comparator_op)

        self.assertRaises(ParseException, parse, "==")
        self.assertRaises(ParseException, parse, "><")
        self.assertRaises(ParseException, parse, "=<")
        self.assertRaises(ParseException, parse, "=>")

    def test_logical_unary(self):
        parse = self.parse_function(self.parser.not_op)

        self.assertEqual(parse("not"), ["not"])
        self.assertEqual(parse("NOT"), ["not"])

    def test_logical_binary(self):
        parse = self.parse_function(self.parser.logical_binary_op)

        self.assertEqual(parse("and"), ["and"])
        self.assertEqual(parse("AND"), ["and"])
        self.assertEqual(parse("or"), ["or"])
        self.assertEqual(parse("OR"), ["or"])

    def test_operand(self):
        parse = self.parse_function(self.parser.operand)

        self.assertEqual(parse('"Secret to life"'), ["'Secret to life'"])
        self.assertEqual(parse("is__"), ["is__"])
        self.assertEqual(parse("1.3337"), ["1.3337"])

    def test_operand_exp(self):
        parse = self.parse_function(self.parser.operand)

        self.assertRaises(ParseException, parse, "=")
        self.assertRaises(ParseException, parse, "Need space!")

    def test_condition(self):
        parse = self.parse_function(self.parser.condition)

        self.assertEqual(parse("2 > 7"), [["2", ">", "7"]])
        self.assertEqual(parse("secret = 3"), [["secret", "=", "3"]])
        self.assertEqual(parse('"name" != "mane"'), [["'name'", "!=", "'mane'"]])

    def test_condition_exp_fragments(self):
        parse = self.parse_function(self.parser.condition)

        self.assertRaises(ParseException, parse, "'Happiness' =")
        self.assertRaises(ParseException, parse, "< you")
        self.assertRaises(ParseException, parse, "<>")

    # Testing a single condition

    def test_conditions_single(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["2", ">", "7"]]
        actual = parse("2 > 7")
        self.assertEqual(actual, expected)

    def test_conditions_par(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["2", ">", "7"]]
        actual = parse("(2 > 7)")
        self.assertEqual(actual, expected)

    def test_conditions_not(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", ["2", ">", "7"]]]
        actual = parse("not (2 > 7)")
        self.assertEqual(actual, expected)

    def test_conditions_and(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "and", ["3", "<>", "3"]]]
        actual = parse("2 > 7 and 3 <> 3")
        self.assertEqual(actual, expected)

    def test_conditions_and_par(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "and", ["3", "<>", "3"]]]
        actual = parse("(2 > 7) and (3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_or(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["2", ">", "7"], "or", ["3", "<>", "3"]]]
        actual = parse("(2 > 7) or (3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["not", ["2", ">", "7"]], "and", ["3", "<>", "3"]]]
        actual = parse("not 2 > 7 and 3 <> 3")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_simple_forced(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", [["2", ">", "7"], "and", ["3", "<>", "3"]]]]
        actual = parse("not (2 > 7 and 3 <> 3)")
        self.assertEqual(actual, expected)

    def test_conditions_binary_unary_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["3", "<>", "3"], "or", ["not", ["2", ">", "7"]]]]
        actual = parse("3 <> 3 or not 2 > 7")
        self.assertEqual(expected, actual)


class TestConditionAttrs(unittest.TestCase):
    def test_conditions_binary_unary(self):
        expected = ["answer"]
        self.assertEqual(expected, get_attribute_references("answer=42"))

    def test_conditions_binary_unary_with_prefix(self):
        expected = ["a.answer"]
        self.assertEqual(expected, get_attribute_references("a.answer=42"))

    def test_conditions_binary_unary_reversed(self):
        expected = ["answer"]
        self.assertEqual(expected, get_attribute_references("42=answer"))

    def test_conditions_binary_unary_no_attr(self):
        self.assertEqual([], get_attribute_references("42=42"))

    def test_conditions_binary_unary_two_attrs(self):
        expected = ["answer", "known"]
        self.assertEqual(expected, get_attribute_references("answer=known"))

    def test_conditions_binary_two_attrs(self):
        expected = ["answer", "known"]
        self.assertEqual(expected, get_attribute_references("answer=known and 42=42"))

    def test_conditions_binary_man_attrs(self):
        expected = ["answer", "known", "answer", "unknown"]
        self.assertEqual(
            expected, get_attribute_references("answer=known and answer=unknown")
        )
